@@ -11,12 +11,11 @@ import matplotlib.pyplot as plt
 from matplotlib.collections import PatchCollection
 from shapely.geometry import Polygon, Point
 from descartes import PolygonPatch
-from geopy.geocoders import Nominatim
 
 from .projection import project_geometry
 from .plot import save_and_show
 from .core import consolidate_subdivide_geometry, get_polygons_coordinates, overpass_request, bbox_from_point, gdf_from_place
-from .utils import log
+from .utils import log, geocode
 
 
 def osm_bldg_download(polygon=None, north=None, south=None, east=None, west=None, 
@@ -41,7 +40,8 @@ def osm_bldg_download(polygon=None, north=None, south=None, east=None, west=None
     memory : int
         server memory allocation size for the query, in bytes. If none, server will use its default allocation size
     max_query_area_size : float
-        max area for any part of the geometry, in the units the geometry is in: any polygon bigger will get divided up for multiple queries to API (default is 50,000 * 50,000 units (ie, 50km x 50km in area, if units are meters))
+        max area for any part of the geometry, in the units the geometry is in: any polygon bigger will get divided 
+        up for multiple queries to API (default is 50,000 * 50,000 units (ie, 50km x 50km in area, if units are meters))
     
     Returns
     -------
@@ -69,7 +69,7 @@ def osm_bldg_download(polygon=None, north=None, south=None, east=None, west=None
     if by_bbox:
         # turn bbox into a polygon and project to local UTM
         polygon = Polygon([(west, south), (east, south), (east, north), (west, north)])
-        geometry_proj, crs_proj = project_geometry(polygon, crs={'init':'epsg:4326'})
+        geometry_proj, crs_proj = project_geometry(polygon)
         
         # subdivide it if it exceeds the max area size (in meters), then project back to lat-long
         geometry_proj_consolidated_subdivided = consolidate_subdivide_geometry(geometry_proj, max_query_area_size=max_query_area_size)
@@ -89,7 +89,7 @@ def osm_bldg_download(polygon=None, north=None, south=None, east=None, west=None
     
     elif by_poly:
         # project to utm, divide polygon up into sub-polygons if area exceeds a max size (in meters), project back to lat-long, then get a list of polygon(s) exterior coordinates
-        geometry_proj, crs_proj = project_geometry(polygon, crs={'init':'epsg:4326'})
+        geometry_proj, crs_proj = project_geometry(polygon)
         geometry_proj_consolidated_subdivided = consolidate_subdivide_geometry(geometry_proj, max_query_area_size=max_query_area_size)
         geometry, crs = project_geometry(geometry_proj_consolidated_subdivided, crs=crs_proj, to_latlong=True)
         polygon_coord_strs = get_polygons_coordinates(geometry)
@@ -201,12 +201,9 @@ def buildings_from_address(address, distance):
     """
     
     # geocode the address string to a (lat, lon) point
-    geolocation = Nominatim().geocode(query=address)
-    try:
-        point = (geolocation.latitude, geolocation.longitude)
-    except:
-        raise Exception('Geocoding failed for address "{}"'.format(address))
-        
+    point = geocode(query=address)
+    
+    # get buildings within distance of this point
     return buildings_from_point(point, distance)
 
 
@@ -222,6 +219,7 @@ def buildings_from_polygon(polygon):
     -------
     GeoDataFrame
     """
+    
     return create_buildings_gdf(polygon=polygon)
     
     
