@@ -25,11 +25,9 @@ from itertools import groupby
 from dateutil import parser as date_parser
 from shapely.geometry import Point, LineString, Polygon, MultiPolygon
 from shapely.ops import unary_union
-from geopy.distance import great_circle, vincenty
-from geopy.geocoders import Nominatim
 
 from . import globals
-from .utils import log, make_str, get_largest_component
+from .utils import log, make_str, get_largest_component, great_circle_vec, get_nearest_node, geocode
 from .simplify import simplify_graph
 from .projection import project_geometry, project_gdf
 from .stats import count_streets_per_node
@@ -37,7 +35,9 @@ from .stats import count_streets_per_node
 
 def save_to_cache(url, response_json):
     """
-    Save a HTTP response json object to the cache. If the request was sent to server via POST instead of GET, 
+    Save an HTTP response json object to the cache. 
+    
+    If the request was sent to server via POST instead of GET, 
     then URL should be a GET-style representation of request. Users should always pass OrderedDicts instead of dicts
     of parameters into request functions, so that the parameters stay in the same order each time, producing the same URL string,
     and thus the same hash. Otherwise the cache will eventually contain multiple saved responses for the same
@@ -85,8 +85,7 @@ def get_from_cache(url):
     
     Returns
     -------
-    dict
-        response_json
+    response_json : dict
     """
     # if the tool is configured to use the cache
     if globals.use_cache:
@@ -166,8 +165,7 @@ def nominatim_request(params, pause_duration=1, timeout=30, error_pause_duration
     
     Returns
     -------
-    dict
-        response_json
+    response_json : dict
     """
     
     # prepare the Nominatim API URL and see if request already exists in the cache
@@ -213,7 +211,7 @@ def nominatim_request(params, pause_duration=1, timeout=30, error_pause_duration
 
 def overpass_request(data, pause_duration=None, timeout=180, error_pause_duration=None):
     """
-    Send a request to the Overpass API via HTTP POST and return the JSON response
+    Send a request to the Overpass API via HTTP POST and return the JSON response.
     
     Parameters
     ----------
@@ -496,7 +494,8 @@ def osm_net_download(polygon=None, north=None, south=None, east=None, west=None,
     memory : int
         server memory allocation size for the query, in bytes. If none, server will use its default allocation size
     max_query_area_size : float
-        max area for any part of the geometry, in the units the geometry is in: any polygon bigger will get divided up for multiple queries to API (default is 50,000 * 50,000 units (ie, 50km x 50km in area, if units are meters))
+        max area for any part of the geometry, in the units the geometry is in: any polygon bigger will get divided up 
+        for multiple queries to API (default is 50,000 * 50,000 units (ie, 50km x 50km in area, if units are meters))
     
     Returns
     -------
@@ -527,7 +526,7 @@ def osm_net_download(polygon=None, north=None, south=None, east=None, west=None,
     if by_bbox:
         # turn bbox into a polygon and project to local UTM
         polygon = Polygon([(west, south), (east, south), (east, north), (west, north)])
-        geometry_proj, crs_proj = project_geometry(polygon, crs={'init':'epsg:4326'})
+        geometry_proj, crs_proj = project_geometry(polygon)
         
         # subdivide it if it exceeds the max area size (in meters), then project back to lat-long
         geometry_proj_consolidated_subdivided = consolidate_subdivide_geometry(geometry_proj, max_query_area_size=max_query_area_size)
@@ -547,7 +546,7 @@ def osm_net_download(polygon=None, north=None, south=None, east=None, west=None,
     
     elif by_poly:
         # project to utm, divide polygon up into sub-polygons if area exceeds a max size (in meters), project back to lat-long, then get a list of polygon(s) exterior coordinates
-        geometry_proj, crs_proj = project_geometry(polygon, crs={'init':'epsg:4326'})
+        geometry_proj, crs_proj = project_geometry(polygon)
         geometry_proj_consolidated_subdivided = consolidate_subdivide_geometry(geometry_proj, max_query_area_size=max_query_area_size)
         geometry, crs = project_geometry(geometry_proj_consolidated_subdivided, crs=crs_proj, to_latlong=True)
         polygon_coord_strs = get_polygons_coordinates(geometry)
@@ -580,8 +579,7 @@ def consolidate_subdivide_geometry(geometry, max_query_area_size):
     
     Returns
     -------
-    Polygon or MultiPolygon
-        geometry
+    geometry : Polygon or MultiPolygon
     """
     
     # let the linear length of the quadrats (with which to subdivide the geometry) be the square root of max area size
@@ -615,8 +613,7 @@ def get_polygons_coordinates(geometry):
     
     Returns
     -------
-    list
-        polygon_coord_strs
+    polygon_coord_strs : list
     """
      
     # extract the exterior coordinates of the geometry to pass to the API later    
@@ -657,6 +654,7 @@ def get_node(element):
     -------
     dict
     """
+    
     node = {}
     node['y'] = element['lat']
     node['x'] = element['lon']
@@ -681,6 +679,7 @@ def get_path(element):
     -------
     dict
     """
+    
     path = {}
     path['osmid'] = element['id']
     
@@ -706,9 +705,9 @@ def parse_osm_nodes_paths(osm_data):
     
     Returns
     -------
-    tuple
-        (nodes, paths)
+    nodes, paths : tuple
     """
+    
     nodes = {}
     paths = {}
     for element in osm_data['elements']:
@@ -724,7 +723,7 @@ def parse_osm_nodes_paths(osm_data):
     
 def remove_isolated_nodes(G):
     """
-    Remove from a graph all the nodes that have no incident edges (ie, node degree = 0)
+    Remove from a graph all the nodes that have no incident edges (ie, node degree = 0).
     
     Parameters
     ----------
@@ -735,6 +734,7 @@ def remove_isolated_nodes(G):
     -------
     networkx multidigraph
     """
+    
     isolated_nodes = [node for node, degree in dict(G.degree()).items() if degree < 1]
     G.remove_nodes_from(isolated_nodes)
     log('Removed {:,} isolated nodes'.format(len(isolated_nodes)))
@@ -780,8 +780,10 @@ def truncate_graph_dist(G, source_node, max_distance=1000, weight='length', reta
     
 def truncate_graph_bbox(G, north, south, east, west, truncate_by_edge=False, retain_all=False):
     """
-    Remove every node in graph that falls outside a bounding box. Needed because overpass returns entire ways that also 
-    include nodes outside the bbox if the way (that is, a way with a single OSM ID) has a node inside the bbox at some point.
+    Remove every node in graph that falls outside a bounding box.
+    
+    Needed because overpass returns entire ways that also include nodes outside the bbox if the way 
+    (that is, a way with a single OSM ID) has a node inside the bbox at some point.
     
     Parameters
     ----------
@@ -803,6 +805,7 @@ def truncate_graph_bbox(G, north, south, east, west, truncate_by_edge=False, ret
     -------
     networkx multidigraph
     """
+    
     start_time = time.time()
     G = G.copy()
     nodes_outside_bbox = []
@@ -994,53 +997,27 @@ def add_edge_lengths(G):
     Returns
     -------
     G : networkx multidigraph
-    """
+    """    
     
     start_time = time.time()
     
-    for u, v, key, data in G.edges(keys=True, data=True):
-        
-        #geopy points are (lat, lon) so that's (y, x) for this great_circle calculation
-        u_point = (G.node[u]['y'], G.node[u]['x'])
-        v_point = (G.node[v]['y'], G.node[v]['x'])
-        edge_length = great_circle(u_point, v_point).m 
-        data['length'] = edge_length
+    # first load all the edges' origin and destination coordinates as a dataframe indexed by u, v, key
+    coords = np.array([[u, v, k, G.node[u]['y'], G.node[u]['x'], G.node[v]['y'], G.node[v]['x']] for u, v, k in G.edges(keys=True)])
+    df_coords = pd.DataFrame(coords, columns=['u', 'v', 'k', 'u_y', 'u_x', 'v_y', 'v_x'])
+    df_coords[['u', 'v', 'k']] = df_coords[['u', 'v', 'k']].astype(np.int64)
+    df_coords = df_coords.set_index(['u', 'v', 'k'])
+
+    # then calculate the great circle distance with the vectorized function
+    gc_distances = great_circle_vec(lat1=df_coords['u_y'], 
+                                    lng1=df_coords['u_x'],
+                                    lat2=df_coords['v_y'], 
+                                    lng2=df_coords['v_x'])
+
+    gc_distances = gc_distances.fillna(value=0)
+    nx.set_edge_attributes(G, 'length', gc_distances.to_dict())
     
     log('Added edge lengths to graph in {:,.2f} seconds'.format(time.time()-start_time))
     return G
-    
-
-def get_nearest_node(G, point, return_dist=False):
-    """
-    Return the graph node nearest to some specified point.
-    
-    Parameters
-    ----------
-    G : networkx multidigraph
-    point : tuple
-        the (lat, lon) point for which we will find the nearest node in the graph
-    return_dist : bool
-        optionally also return the distance between the point and the nearest node
-    
-    Returns
-    -------
-    networkx multidigraph or tuple
-        multidigraph or optionally (multidigraph, float)
-    """
-    start_time = time.time()
-    nodes = G.nodes(data=True)
-    
-    # the nearest node is the one that minimizes great circle distance between its coordinates and the passed-in point
-    # geopy points are (lat, lon) so that's (y, x)
-    nearest_node = min(nodes, key=lambda node: great_circle((node[1]['y'], node[1]['x']), point).m)
-    log('Found nearest node ({}) to point {} in {:,.2f} seconds'.format(nearest_node[0], point, time.time()-start_time))
-    
-    if return_dist:
-        # if caller requested return_dist, calculate the great circle distance between the point and the nearest node and return it as well
-        distance = great_circle((nearest_node[1]['y'], nearest_node[1]['x']), point).m
-        return nearest_node[0], distance
-    else:
-        return nearest_node[0]
 
         
 def add_path(G, data, one_way):
@@ -1059,6 +1036,7 @@ def add_path(G, data, one_way):
     -------
     None
     """
+    
     # extract the ordered list of nodes from this path element, then delete it so we don't add it as an attribute to the edge later
     path_nodes = data['nodes']
     del data['nodes']
@@ -1093,6 +1071,7 @@ def add_paths(G, paths, network_type):
     -------
     None
     """
+    
     # the list of values OSM uses in its 'oneway' tag to denote True
     osm_oneway_values = ['yes', 'true', '1', '-1']
     
@@ -1134,6 +1113,7 @@ def create_graph(response_jsons, name='unnamed', retain_all=False, network_type=
     -------
     networkx multidigraph
     """
+    
     log('Creating networkx graph from downloaded OSM data...')
     start_time = time.time()
     
@@ -1178,7 +1158,7 @@ def create_graph(response_jsons, name='unnamed', retain_all=False, network_type=
     
 def bbox_from_point(point, distance=1000, project_utm=False):
     """
-    Create a bounding box some distance in each direction (north, south, east, and west) from some (lat, lon) point.
+    Create a bounding box some distance in each direction (north, south, east, and west) from some (lat, lng) point.
     
     Parameters
     ----------
@@ -1191,21 +1171,21 @@ def bbox_from_point(point, distance=1000, project_utm=False):
     
     Returns
     -------
-    tuple
-        north, south, east, west
+    north, south, east, west : tuple
     """
-    north = vincenty(meters=distance).destination(point, bearing=0).latitude
-    south = vincenty(meters=distance).destination(point, bearing=180).latitude
-    east = vincenty(meters=distance).destination(point, bearing=90).longitude
-    west = vincenty(meters=distance).destination(point, bearing=270).longitude
+    
+    # reverse the order of the (lat,lng) point so it is (x,y) for shapely, then project to UTM and buffer in meters
+    lat, lng = point
+    point_proj, crs_proj = project_geometry(Point((lng, lat)))
+    buffer_proj = point_proj.buffer(distance)
     
     if project_utm:
-        # if user requests UTM coords, create a polygon from these coords, then project to UTM, extract the bounds, and return
-        bbox_utm, crs_utm = project_geometry(geometry=Polygon([(west, north), (west, south), (east, south), (east, north), (west, north)]), 
-                                             crs={'init':'epsg:4326'})
-        west, south, east, north = bbox_utm.bounds
+        west, south, east, north = buffer_proj.bounds
         log('Created bounding box {} meters in each direction from {} and projected it: {},{},{},{}'.format(distance, point, north, south, east, west))
     else:
+        # if project_utm is False, project back to lat-long then get the bounding coordinates
+        buffer_latlong, crs_latlong = project_geometry(buffer_proj, crs=crs_proj, to_latlong=True)
+        west, south, east, north = buffer_latlong.bounds
         log('Created bounding box {} meters in each direction from {}: {},{},{},{}'.format(distance, point, north, south, east, west))
         
     return north, south, east, west
@@ -1255,7 +1235,7 @@ def graph_from_bbox(north, south, east, west, network_type='all_private', simpli
         # create a new buffered bbox 0.5km around the desired one
         buffer_dist = 500
         polygon = Polygon([(west, north), (west, south), (east, south), (east, north)])
-        polygon_utm, crs_utm = project_geometry(geometry=polygon, crs={'init':'epsg:4326'})
+        polygon_utm, crs_utm = project_geometry(geometry=polygon)
         polygon_proj_buff = polygon_utm.buffer(buffer_dist)
         polygon_buff, crs = project_geometry(geometry=polygon_proj_buff, crs=crs_utm, to_latlong=True)
         west_buffered, south_buffered, east_buffered, north_buffered = polygon_buff.bounds
@@ -1304,7 +1284,8 @@ def graph_from_point(center_point, distance=1000, distance_type='bbox', network_
     distance : int
         retain only those nodes within this many meters of the center of the graph
     distance_type : string
-        {'network', 'bbox'} if 'bbox', retain only those nodes within a bounding box of the distance parameter. if 'network', retain only those nodes within some network distance from the center-most node.
+        {'network', 'bbox'} if 'bbox', retain only those nodes within a bounding box of the distance parameter.
+        if 'network', retain only those nodes within some network distance from the center-most node.
     network_type : string
         what type of street network to get
     simplify : bool
@@ -1330,28 +1311,21 @@ def graph_from_point(center_point, distance=1000, distance_type='bbox', network_
     networkx multidigraph
     """
     
+    if distance_type not in ['bbox', 'network']:
+        raise ValueError('distance_type must be "bbox" or "network"')
+    
     # create a bounding box from the center point and the distance in each direction
     north, south, east, west = bbox_from_point(center_point, distance)
-    if distance_type == 'bbox':
-        # if the network distance_type is bbox, create a graph from the bounding box
-        G = graph_from_bbox(north, south, east, west, network_type=network_type, simplify=simplify, retain_all=retain_all, 
-                            truncate_by_edge=truncate_by_edge, name=name, timeout=timeout, memory=memory, max_query_area_size=max_query_area_size, clean_periphery=clean_periphery)
-    elif distance_type == 'network':
-        # if the network distance_type is network, create a graph from the bounding box but do not simplify it yet (only simplify a graph after all truncation is performed! otherwise you get weird artifacts)
-        G = graph_from_bbox(north, south, east, west, network_type=network_type, simplify=False, retain_all=retain_all, 
-                            truncate_by_edge=truncate_by_edge, name=name, timeout=timeout, memory=memory, max_query_area_size=max_query_area_size, clean_periphery=clean_periphery)
-        
-        # next find the node in the graph nearest to the center point, and truncate the graph by network distance from this node
+    
+    # create a graph from the bounding box
+    G = graph_from_bbox(north, south, east, west, network_type=network_type, simplify=simplify, retain_all=retain_all, 
+                        truncate_by_edge=truncate_by_edge, name=name, timeout=timeout, memory=memory, 
+                        max_query_area_size=max_query_area_size, clean_periphery=clean_periphery)
+
+    # if the network distance_type is network, find the node in the graph nearest to the center point, and truncate the graph by network distance from this node
+    if distance_type == 'network':
         centermost_node = get_nearest_node(G, center_point)
         G = truncate_graph_dist(G, centermost_node, max_distance=distance)
-        
-        # simplify the graph topology as the last step. don't truncate after simplifying or you may have simplified out to an endpoint
-        # beyond the truncation distance, in which case you will then strip out your entire edge
-        # that's why simplify=False above, so we didn't do it before the truncate_graph_dist() call 2 lines after it
-        if simplify:
-            G = simplify_graph(G)
-    else:
-        raise ValueError('distance_type must be "bbox" or "network"')
     
     log('graph_from_point() returning graph with {:,} nodes and {:,} edges'.format(len(list(G.nodes())), len(list(G.edges()))))
     return G
@@ -1369,7 +1343,8 @@ def graph_from_address(address, distance=1000, distance_type='bbox', network_typ
     distance : int
         retain only those nodes within this many meters of the center of the graph
     distance_type : string
-        {'network', 'bbox'} if 'bbox', retain only those nodes within a bounding box of the distance parameter. if 'network', retain only those nodes within some network distance from the center-most node.
+        {'network', 'bbox'} if 'bbox', retain only those nodes within a bounding box of the distance parameter.
+        if 'network', retain only those nodes within some network distance from the center-most node.
     network_type : string
         what type of street network to get
     simplify : bool
@@ -1399,11 +1374,7 @@ def graph_from_address(address, distance=1000, distance_type='bbox', network_typ
     """
     
     # geocode the address string to a (lat, lon) point
-    geolocation = Nominatim().geocode(query=address)
-    try:
-        point = (geolocation.latitude, geolocation.longitude)
-    except:
-        raise Exception('Geocoding failed for address "{}"'.format(address))
+    point = geocode(query=address)
     
     # then create a graph from this point
     G = graph_from_point(point, distance, distance_type, network_type=network_type, simplify=simplify, retain_all=retain_all, 
@@ -1459,7 +1430,7 @@ def graph_from_polygon(polygon, network_type='all_private', simplify=True, retai
     if clean_periphery and simplify:
         # create a new buffered polygon 0.5km around the desired one
         buffer_dist = 500
-        polygon_utm, crs_utm = project_geometry(geometry=polygon, crs={'init':'epsg:4326'})
+        polygon_utm, crs_utm = project_geometry(geometry=polygon)
         polygon_proj_buff = polygon_utm.buffer(buffer_dist)
         polygon_buffered, crs = project_geometry(geometry=polygon_proj_buff, crs=crs_utm, to_latlong=True)
         
